/******************************
 * 共有メモリイメージのフォーマットを列挙
 ******************************/

#include <stdio.h>
#include <string.h>
#include <wayland-client.h>

struct wl_shm *g_shm;


//---------------------
// wl_shm
//---------------------


static void _shm_format(void *data,struct wl_shm *shm,uint32_t format)
{
	if(format == WL_SHM_FORMAT_ARGB8888)
		printf("%u: ARGB8888\n", format);
	else if(format == WL_SHM_FORMAT_XRGB8888)
		printf("%u: XRGB8888\n", format);
	else
	{
		printf("0x%08X: '%c%c%c%c'\n",
			format,
			(char)(format >> 24),
			(char)(format >> 16),
			(char)(format >> 8),
			(char)format);
	}
}

static const struct wl_shm_listener g_shm_listener = { _shm_format };


//---------------------
// wl_registry
//---------------------


static void _registry_global(
	void *data,struct wl_registry *registry,
	uint32_t id,const char *interface,uint32_t version)
{
	if(strcmp(interface, "wl_shm") == 0)
	{
		//wl_shm をバインド
		
		g_shm = wl_registry_bind(registry, id, &wl_shm_interface, 1);

		wl_shm_add_listener(g_shm, &g_shm_listener, NULL);
	}
}

static void _registry_global_remove(void *data,struct wl_registry *registry,uint32_t id)
{

}

static const struct wl_registry_listener g_reg_listener = {
	_registry_global, _registry_global_remove
};


//---------------------


int main(void)
{
	struct wl_display *display;
	struct wl_registry *reg;

	display = wl_display_connect(NULL);
	if(!display) return 1;

	//wl_registry

	reg = wl_display_get_registry(display);
	wl_registry_add_listener(reg, &g_reg_listener, NULL);

	//wl_registry のイベントを待つ

	wl_display_roundtrip(display);

	//wl_shm のイベントを待つ

	wl_display_roundtrip(display);

	//

	wl_shm_destroy(g_shm);
	wl_registry_destroy(reg);

	wl_display_disconnect(display);

	return 0;
}
