/******************************
 * ウィンドウの表示
 ******************************/

#include <stdio.h>
#include <wayland-client.h>

#include "client.h"
#include "imagebuf.h"


//---------------

typedef struct
{
	Client *client;
	Window *win;
	int count;
}callback_data;

//---------------


/* イメージ描画
 *
 * 黒→赤のグラデーション */

static void draw_image(ImageBuf *p,int alpha)
{
	uint8_t *pd = (uint8_t *)p->data,val;
	int ix,iy;

	for(iy = p->height; iy > 0; iy--)
	{
		for(ix = p->width, val = 0; ix > 0; ix--, pd += 4)
		{
			//B-G-R-A (Little Endian)
		
			pd[0] = pd[1] = 0;
			pd[2] = val++;
			pd[3] = alpha;
		}
	}
}


//--------------------------
// wl_callback
//--------------------------


static void _callback_done(void *data,struct wl_callback *callback,uint32_t time_ms);

static const struct wl_callback_listener g_callback_listener = {
	_callback_done
};

/* done イベント */

void _callback_done(void *data,struct wl_callback *callback,uint32_t time_ms)
{
	callback_data *p = (callback_data *)data;

	wl_callback_destroy(callback);

	p->count++;

	printf("count: %d\r", p->count);
	fflush(stdout);

	if(p->count == 256)
		//イベントループを終了させる
		p->client->finish_loop = 1;
	else
	{
		//---- 更新

		draw_image(p->win->img, p->count);

		//新しいコールバック

		callback = wl_surface_frame(p->win->surface);

		wl_callback_add_listener(callback, &g_callback_listener, data);

		//適用

		Window_update(p->win);

		/* Window_update() の代わりに以下を使うと、
		 * 画像のアルファ値が適用されない (常に不透明) */

		//Window_updateOpaque(p->win);
	}
}


//---------------


int main(void)
{
	Client *client;
	Window *window;
	struct wl_callback *callback;
	callback_data dat;

	//初期化

	client = Client_new(0);

	Client_init(client);

	//ウィンドウ作成
	/* wl_surface
	 * wl_shell_surface */

	window = Window_create(client, 256, 256, NULL);

	//アニメ用コールバック

	dat.client = client;
	dat.win = window;
	dat.count = 0;

	callback = wl_surface_frame(window->surface);

	wl_callback_add_listener(callback, &g_callback_listener, &dat);

	//最初の画面を描画&更新

	draw_image(window->img, 0);

	Window_update(window);

	//イベントループ

	Client_loop_simple(client);

	//

	Window_destroy(window);

	Client_destroy(client);

	return 0;
}
