/******************************
 * クリップボード
 ******************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define __USE_GNU
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <poll.h>

#include <linux/input.h>

#include <wayland-client.h>

#include "client.h"
#include "imagebuf.h"


//-------------

typedef struct
{
	Client b;
	struct pollfd poll[2];

	struct wl_data_device_manager *data_device_manager;
	struct wl_data_device *data_device;
	struct wl_data_source *data_source;
	struct wl_data_offer *data_offer;

	uint32_t data_device_manager_ver;

	char *copy_text,	//コピー文字列
		*recv_buf;		//受信バッファ
	int text_size,		//コピー文字列のサイズ
		recv_cursize,	//受信バッファ、現在の読み込みサイズ
		recv_bufsize;	//受信バッファ確保サイズ
}ClientEx;

//-------------


/* ソースデータ解放 */

static void _source_release(ClientEx *p)
{
	if(p->data_source)
	{
		wl_data_source_destroy(p->data_source);
		p->data_source = NULL;
	}

	if(p->copy_text)
	{
		free(p->copy_text);
		p->copy_text = NULL;
	}
}


//-----------------------
// wl_data_source
//-----------------------


static void _data_source_target(void *data,
	struct wl_data_source *source, const char *mime_type)
{
	//D&D
}

/* データが要求された時 */

static void _data_source_send(void *data, struct wl_data_source *source,
	const char *mime_type, int32_t fd)
{
	ClientEx *p = (ClientEx *)data;

	if(!mime_type)
		close(fd);
	else
	{
		printf("wl_data_source # send | mime_type:\"%s\", fd:%d\n", mime_type, fd);

		//fd に書き込んで、送る

		write(fd, p->copy_text, p->text_size);
		close(fd);
	}
}

/* データソースが置き換わった時 */

static void _data_source_cancelled(void *data, struct wl_data_source *source)
{
	printf("wl_data_source # cancelled | source:%p\n", source);

	_source_release((ClientEx *)data);
}

void _data_source_dnd_drop_performed(void *data, struct wl_data_source *source)
{
	//D&D
}

void _data_source_dnd_finished(void *data, struct wl_data_source *source)
{
	//D&D
}

void _data_source_action(void *data, struct wl_data_source *source, uint32_t dnd_action)
{
	//D&D
}

static const struct wl_data_source_listener g_data_source_listener = {
	_data_source_target, _data_source_send, _data_source_cancelled,
	_data_source_dnd_drop_performed, _data_source_dnd_finished,
	_data_source_action
};


//-----------------------
// wl_data_offer
//-----------------------


/* 提供されている MIME タイプが通知される */

static void _data_offer_offer(void *data, struct wl_data_offer *offer, const char *type)
{
	printf("wl_data_offer # offer | type:\"%s\"\n", type);
}

static void _data_offer_source_actions(void *data,
	struct wl_data_offer *offer, uint32_t source_actions)
{
	//D&D
}

static void _data_offer_action(void *data, struct wl_data_offer *offer, uint32_t dnd_action)
{
	//D&D
}

static const struct wl_data_offer_listener g_data_offer_listener = {
	_data_offer_offer, _data_offer_source_actions, _data_offer_action
};


//-----------------------
// wl_data_device
//-----------------------


static void _device_data_offer(void *data,
	struct wl_data_device *data_device, struct wl_data_offer *offer)
{
	printf("wl_data_device # data_offer | offer:%p\n", offer);

	wl_data_offer_add_listener(offer, &g_data_offer_listener, data);
}

static void _device_enter(void *data, struct wl_data_device *data_device, uint32_t serial,
	struct wl_surface *surface, wl_fixed_t x, wl_fixed_t y, struct wl_data_offer *offer)
{
	//D&D
}

static void _device_leave(void *data, struct wl_data_device *data_device)
{
	//D&D
}

static void _device_motion(void *data, struct wl_data_device *data_device,
	uint32_t time, wl_fixed_t x, wl_fixed_t y)
{
	//D&D
}

static void _device_drop(void *data, struct wl_data_device *data_device)
{
	//D&D
}

/* クリップボード用の wl_data_offer が送られてくる */

static void _device_selection(void *data,
	struct wl_data_device *data_device, struct wl_data_offer *offer)
{
	ClientEx *p = (ClientEx *)data;

	printf("wl_data_device # selection | offer:%p\n", offer);

	//前回の data_offer は破棄する

	if(p->data_offer)
		wl_data_offer_destroy(p->data_offer);

	//offer を保持
	//(クリップボードにデータがない場合は NULL となる)

	p->data_offer = offer;
}

static const struct wl_data_device_listener g_data_device_listener = {
	_device_data_offer, _device_enter, _device_leave,
	_device_motion, _device_drop, _device_selection
};


//-----------------------
// コピー/貼り付け処理
//-----------------------


/* コピー */

static void _copy_text(ClientEx *p,const char *text,uint32_t serial)
{
	printf("-- copy text: \"%s\"\n", text);

	//前のソースを解放

	_source_release(p);

	//内部データに文字列をコピー

	p->text_size = strlen(text);

	p->copy_text = (char *)malloc(p->text_size);

	memcpy(p->copy_text, text, p->text_size);

	//このクライアントがデータを持っていることを通知する

	p->data_source = wl_data_device_manager_create_data_source(p->data_device_manager);

	wl_data_source_offer(p->data_source, "text/plain;charset=utf-8");
	wl_data_source_offer(p->data_source, "UTF8_STRING");

	wl_data_source_add_listener(p->data_source, &g_data_source_listener, p);

	wl_data_device_set_selection(p->data_device, p->data_source, serial);
}

/* 貼り付け */

static void _paste(ClientEx *p)
{
	int fd[2];

	if(!p->data_offer)
	{
		printf("-- paste: no data\n");
		return;
	}

	//現在、データを受信中

	if(p->poll[1].fd != -1) return;

	//データの受信を要求

	if(pipe2(fd, O_CLOEXEC) == -1) return;

	wl_data_offer_receive(p->data_offer, "text/plain;charset=utf-8", fd[1]);
	close(fd[1]);

	p->poll[1].fd = fd[0];
	p->recv_cursize = 0;
}

/* 貼り付け時のデータ受信 */

static void _data_recv_handle(ClientEx *p,int events)
{
	int fd,ret = 0;

	fd = p->poll[1].fd;

	printf("@ poll | ");
	if(events & POLLIN) printf("POLLIN ");
	if(events & POLLHUP) printf("POLLHUP ");
	printf("\n");

	//受信するデータがある場合

	if(events & POLLIN)
	{
		//最後に NULL 文字を追加するので、読み込みサイズは 1 byte 減らす

		ret = read(fd, p->recv_buf + p->recv_cursize,
			p->recv_bufsize - p->recv_cursize - 1);

		p->recv_cursize += ret;

		printf("@ recv | read:%d byte, buf:%d/%d byte\n", (int)ret, p->recv_cursize, p->recv_bufsize);
	}

	//

	if(ret == 0)
	{
		//受信データがなくなったら終了

		close(fd);
		p->poll[1].fd = -1;

		//NULL 文字を追加
		p->recv_buf[p->recv_cursize] = 0;

		printf("<paste text>\n");
		puts(p->recv_buf);
	}
	else
	{
		//データが続く場合、常に 1 KB 以上の空き容量を作っておく

		if(p->recv_bufsize - p->recv_cursize < 1024)
		{
			p->recv_bufsize += 1024;
			p->recv_buf = (char *)realloc(p->recv_buf, p->recv_bufsize);
		}
	}
}


//-----------------------
// wl_keyboard
//-----------------------


static void _keyboard_keymap(void *data, struct wl_keyboard *keyboard,
	uint32_t format, int32_t fd, uint32_t size)
{
	//キーマップ処理はしない
}

static void _keyboard_enter(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, struct wl_surface *surface, struct wl_array *keys)
{

}

static void _keyboard_leave(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, struct wl_surface *surface)
{

}

/* キー */

static void _keyboard_key(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, uint32_t time, uint32_t key, uint32_t state)
{
	ClientEx *p = (ClientEx *)data;
	char m[64];

	if(state != WL_KEYBOARD_KEY_STATE_PRESSED) return;

	switch(key)
	{
		//コピー
		case KEY_C:
			snprintf(m, 64, "copytext-%u", serial);
			_copy_text(p, m, serial);
			break;
		//貼り付け
		case KEY_P:
			_paste(p);
			break;
		//クリップボード解除
		case KEY_R:
			if(p->data_source)
			{
				wl_data_device_set_selection(p->data_device, NULL, serial);
				_source_release(p);
			}
			break;
		//ESC キーで終了
		case KEY_ESC:
			p->b.finish_loop = 1;
			break;
	}
}

static void _keyboard_modifiers(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, uint32_t mods_depressed, uint32_t mods_latched, uint32_t mods_locked, uint32_t group)
{

}

static void _keyboard_repeat_info(void *data, struct wl_keyboard *keyboard,
	int32_t rate, int32_t delay)
{

}

static const struct wl_keyboard_listener g_keyboard_listener = {
	_keyboard_keymap, _keyboard_enter, _keyboard_leave, _keyboard_key,
	_keyboard_modifiers, _keyboard_repeat_info
};


//-----------------------
// wl_registry
//-----------------------


static void _registry_global(
	void *data,struct wl_registry *reg,uint32_t id,const char *name,uint32_t ver)
{
	ClientEx *p = (ClientEx *)data;

	if(strcmp(name, "wl_data_device_manager") == 0)
	{
		if(ver >= 3) ver = 3;

		p->data_device_manager = wl_registry_bind(reg, id, &wl_data_device_manager_interface, ver);
		p->data_device_manager_ver = ver;
	}
}


//--------------------


/* イベントループ */

static void _event_loop(ClientEx *p)
{
	//display イベント用

	p->poll[0].fd = wl_display_get_fd(p->b.display);
	p->poll[0].events = POLLIN;

	//貼り付けデータ取得用
	//(初期状態ではデータがないため、fd = -1。
	// fd が負の値だと poll() 時に無視される)

	p->poll[1].fd = -1;
	p->poll[1].events = POLLIN;

	//ループ

	while(!p->b.finish_loop)
	{
		wl_display_flush(p->b.display);

		if(poll(p->poll, 2, -1) < 0) break;

		//display にイベントがある

		if(p->poll[0].revents & POLLIN)
			wl_display_dispatch(p->b.display);

		//データ受信

		if(p->poll[1].revents)
			_data_recv_handle(p, p->poll[1].revents);
	}
}

/* ClientEx 破棄 */

static void _clientex_destroy(Client *cl)
{
	ClientEx *p = (ClientEx *)cl;

	free(p->recv_buf);

	//wl_data_offer

	if(p->data_offer)
		wl_data_offer_destroy(p->data_offer);

	//ソースデータ

	_source_release(p);

	//wl_data_device

	if(p->data_device_manager_ver >= WL_DATA_DEVICE_RELEASE_SINCE_VERSION)
		wl_data_device_release(p->data_device);
	else
		wl_data_device_destroy(p->data_device);

	//wl_data_device_manager

	wl_data_device_manager_destroy(p->data_device_manager);
}


//-------------------


int main(void)
{
	ClientEx *p;
	Window *win;

	p = (ClientEx *)Client_new(sizeof(ClientEx));

	p->b.destroy = _clientex_destroy;
	p->b.init_flags = INIT_FLAGS_SEAT | INIT_FLAGS_KEYBOARD;
	p->b.registry_global = _registry_global;
	p->b.keyboard_listener = &g_keyboard_listener;

	Client_init(CLIENT(p));

	//wl_data_device

	p->data_device = wl_data_device_manager_get_data_device(
		p->data_device_manager, p->b.seat);

	wl_data_device_add_listener(p->data_device, &g_data_device_listener, p);

	//貼り付け時用の受信バッファ

	p->recv_buf = (char *)malloc(1024);
	p->recv_bufsize = 1024;

	//ウィンドウ

	win = Window_create(CLIENT(p), 256, 256, NULL);

	ImageBuf_fill(win->img, 0xffff0000);
	Window_update(win);

	//

	_event_loop(p);

	//解放

	Window_destroy(win);

	Client_destroy(CLIENT(p));

	return 0;
}
